@@ -1,4 +1,3 @@
-# encoding: utf-8
 """A simple configuration system."""
 
 # Copyright (c) IPython Development Team.
@@ -6,18 +5,19 @@
 
 import argparse
 import copy
-import logging
 import os
 import re
 import sys
 import json
-from ast import literal_eval
+import warnings
+
+from ..utils import cast_unicode
 
 from ipython_genutils.path import filefind
-from ipython_genutils import py3compat
 from ipython_genutils.encoding import DEFAULT_ENCODING
-from six import text_type
-from traitlets.traitlets import HasTraits, List, Any
+from traitlets.traitlets import (
+    HasTraits, Container, List, Dict, Any, Undefined,
+)
 
 #-----------------------------------------------------------------------------
 # Exceptions
@@ -46,6 +46,18 @@ class ArgumentError(ConfigLoaderError):
 # to do.  So we override the print_help method with one that defaults to
 # stdout and use our class instead.
 
+
+class _Sentinel:
+    def __repr__(self):
+        return "<Sentinel deprecated>"
+
+    def __str__(self):
+        return "<deprecated>"
+
+
+_deprecated = _Sentinel()
+
+
 class ArgumentParser(argparse.ArgumentParser):
     """Simple argparse subclass that prints help to stdout by default."""
 
@@ -60,56 +72,102 @@ class ArgumentParser(argparse.ArgumentParser):
 # Config class for holding config information
 #-----------------------------------------------------------------------------
 
+def execfile(fname, glob):
+    with open(fname, 'rb') as f:
+        exec(compile(f.read(), fname, 'exec'), glob, glob)
+
 class LazyConfigValue(HasTraits):
     """Proxy object for exposing methods on configurable containers
-    
+
+    These methods allow appending/extending/updating
+    to add to non-empty defaults instead of clobbering them.
+
     Exposes:
-    
+
     - append, extend, insert on lists
     - update on dicts
     - update, add on sets
     """
-    
+
     _value = None
-    
+
     # list methods
     _extend = List()
     _prepend = List()
-    
+    _inserts = List()
+
     def append(self, obj):
+        """Append an item to a List"""
         self._extend.append(obj)
-    
+
     def extend(self, other):
+        """Extend a list"""
         self._extend.extend(other)
-    
+
     def prepend(self, other):
         """like list.extend, but for the front"""
         self._prepend[:0] = other
-    
-    _inserts = List()
+
+
+    def merge_into(self, other):
+        """
+        Merge with another earlier LazyConfigValue or an earlier container.
+        This is useful when having global system-wide configuration files.
+
+        Self is expected to have higher precedence.
+
+        Parameters
+        ----------
+        other : LazyConfigValue or container
+
+        Returns
+        -------
+        LazyConfigValue
+            if ``other`` is also lazy, a reified container otherwise.
+        """
+        if isinstance(other, LazyConfigValue):
+            other._extend.extend(self._extend)
+            self._extend = other._extend
+
+            self._prepend.extend(other._prepend)
+
+            other._inserts.extend(self._inserts)
+            self._inserts = other._inserts
+
+            if self._update:
+                other.update(self._update)
+                self._update = other._update
+            return self
+        else:
+            # other is a container, reify now.
+            return self.get_value(other)
+
     def insert(self, index, other):
         if not isinstance(index, int):
             raise TypeError("An integer is required")
         self._inserts.append((index, other))
-    
+
     # dict methods
     # update is used for both dict and set
     _update = Any()
+
     def update(self, other):
+        """Update either a set or dict"""
         if self._update is None:
             if isinstance(other, dict):
                 self._update = {}
             else:
                 self._update = set()
         self._update.update(other)
-    
+
     # set methods
     def add(self, obj):
+        """Add an item to a set"""
         self.update({obj})
-    
+
     def get_value(self, initial):
         """construct the value from the initial one
-        
+
         after applying any insert / extend / update changes
         """
         if self._value is not None:
@@ -120,7 +178,7 @@ class LazyConfigValue(HasTraits):
                 value.insert(idx, obj)
             value[:0] = self._prepend
             value.extend(self._extend)
-        
+
         elif isinstance(value, dict):
             if self._update:
                 value.update(self._update)
@@ -129,10 +187,10 @@ class LazyConfigValue(HasTraits):
                 value.update(self._update)
         self._value = value
         return value
-    
+
     def to_dict(self):
         """return JSONable dict form of my data
-        
+
         Currently update as dict or set, extend, prepend as lists, and inserts as list of tuples.
         """
         d = {}
@@ -146,6 +204,12 @@ class LazyConfigValue(HasTraits):
             d['inserts'] = self._inserts
         return d
 
+    def __repr__(self):
+        if self._value is not None:
+            return "<%s value=%r>" % (self.__class__.__name__, self._value)
+        else:
+            return "<%s %r>" % (self.__class__.__name__, self.to_dict())
+
 
 def _is_section_key(key):
     """Is a Config key a section name (does it start with a capital)?"""
@@ -156,15 +220,26 @@ def _is_section_key(key):
 
 
 class Config(dict):
-    """An attribute based dict that can do smart merges."""
+    """An attribute-based dict that can do smart merges.
+
+    Accessing a field on a config object for the first time populates the key
+    with either a nested Config object for keys starting with capitals
+    or :class:`.LazyConfigValue` for lowercase keys,
+    allowing quick assignments such as::
+
+        c = Config()
+        c.Class.int_trait = 5
+        c.Class.list_trait.append("x")
+
+    """
 
     def __init__(self, *args, **kwds):
         dict.__init__(self, *args, **kwds)
         self._ensure_subconfig()
-    
+
     def _ensure_subconfig(self):
         """ensure that sub-dicts that should be Config objects are
-        
+
         casts dicts that are under section keys to Config objects,
         which is necessary for constructing Config objects from dict literals.
         """
@@ -174,11 +249,11 @@ class Config(dict):
                     and isinstance(obj, dict) \
                     and not isinstance(obj, Config):
                 setattr(self, key, Config(obj))
-    
+
     def _merge(self, other):
         """deprecated alias, use Config.merge()"""
         self.merge(other)
-    
+
     def merge(self, other):
         """merge another config object into this one"""
         to_update = {}
@@ -189,18 +264,20 @@ class Config(dict):
                 if isinstance(v, Config) and isinstance(self[k], Config):
                     # Recursively merge common sub Configs
                     self[k].merge(v)
+                elif isinstance(v, LazyConfigValue):
+                    self[k] = v.merge_into(self[k])
                 else:
                     # Plain updates for non-Configs
                     to_update[k] = v
 
         self.update(to_update)
-    
+
     def collisions(self, other):
         """Check for collisions between two config objects.
-        
+
         Returns a dict of the form {"Class": {"trait": "collision message"}}`,
         indicating which values have been ignored.
-        
+
         An empty dict indicates no collisions.
         """
         collisions = {}
@@ -214,7 +291,7 @@ class Config(dict):
                     collisions.setdefault(section, {})
                     collisions[section][key] = "%r ignored, using %r" % (mine[key], theirs[key])
         return collisions
-    
+
     def __contains__(self, key):
         # allow nested contains of the form `"Section.key" in config`
         if '.' in key:
@@ -222,15 +299,15 @@ class Config(dict):
             if first not in self:
                 return False
             return remainder in self[first]
-        
+
         return super(Config, self).__contains__(key)
-    
+
     # .has_key is deprecated for dictionaries.
     has_key = __contains__
-    
+
     def _has_section(self, key):
         return _is_section_key(key) and key in self
-    
+
     def copy(self):
         return type(self)(dict.copy(self))
 
@@ -248,7 +325,7 @@ class Config(dict):
                 value = copy.copy(value)
             new_config[key] = value
         return new_config
-    
+
     def __getitem__(self, key):
         try:
             return dict.__getitem__(self, key)
@@ -297,6 +374,89 @@ class Config(dict):
             raise AttributeError(e)
 
 
+class DeferredConfig:
+    """Class for deferred-evaluation of config from CLI"""
+    pass
+
+    def get_value(self, trait):
+        raise NotImplementedError("Implement in subclasses")
+
+    def _super_repr(self):
+        # explicitly call super on direct parent
+        return super(self.__class__, self).__repr__()
+
+
+class DeferredConfigString(str, DeferredConfig):
+    """Config value for loading config from a string
+
+    Interpretation is deferred until it is loaded into the trait.
+
+    Subclass of str for backward compatibility.
+
+    This class is only used for values that are not listed
+    in the configurable classes.
+
+    When config is loaded, `trait.from_string` will be used.
+
+    If an error is raised in `.from_string`,
+    the original string is returned.
+
+    .. versionadded:: 5.0
+    """
+    def get_value(self, trait):
+        """Get the value stored in this string"""
+        s = str(self)
+        try:
+            return trait.from_string(s)
+        except Exception:
+            # exception casting from string,
+            # let the original string lie.
+            # this will raise a more informative error when config is loaded.
+            return s
+
+    def __repr__(self):
+        return '%s(%s)' % (self.__class__.__name__, self._super_repr())
+
+
+class DeferredConfigList(list, DeferredConfig):
+    """Config value for loading config from a list of strings
+
+    Interpretation is deferred until it is loaded into the trait.
+
+    This class is only used for values that are not listed
+    in the configurable classes.
+
+    When config is loaded, `trait.from_string_list` will be used.
+
+    If an error is raised in `.from_string_list`,
+    the original string list is returned.
+
+    .. versionadded:: 5.0
+    """
+    def get_value(self, trait):
+        """Get the value stored in this string"""
+        if hasattr(trait, "from_string_list"):
+            src = list(self)
+            cast = trait.from_string_list
+        else:
+            # only allow one item
+            if len(self) > 1:
+                raise ValueError(f"{trait.name} only accepts one value, got {len(self)}: {list(self)}")
+            src = self[0]
+            cast = trait.from_string
+
+        try:
+            return cast(src)
+        except Exception:
+            # exception casting from string,
+            # let the original value lie.
+            # this will raise a more informative error when config is loaded.
+            return src
+
+    def __repr__(self):
+        return '%s(%s)' % (self.__class__.__name__, self._super_repr())
+
+
 #-----------------------------------------------------------------------------
 # Config loading classes
 #-----------------------------------------------------------------------------
@@ -325,12 +485,11 @@ class ConfigLoader(object):
         """A base class for config loaders.
 
         log : instance of :class:`logging.Logger` to use.
-              By default loger of :meth:`traitlets.config.application.Application.instance()`
+              By default logger of :meth:`traitlets.config.application.Application.instance()`
               will be used
 
         Examples
         --------
-
         >>> cl = ConfigLoader()
         >>> config = cl.load_config()
         >>> config
@@ -456,12 +615,12 @@ class PyFileConfigLoader(FileConfigLoader):
             raise ConfigFileNotFound(str(e))
         self._read_file_as_dict()
         return self.config
-    
+
     def load_subconfig(self, fname, path=None):
         """Injected into config file namespace as load_subconfig"""
         if path is None:
             path = self.path
-        
+
         loader = self.__class__(fname, path)
         try:
             sub_config = loader.load_config()
@@ -471,22 +630,22 @@ class PyFileConfigLoader(FileConfigLoader):
             pass
         else:
             self.config.merge(sub_config)
-    
+
     def _read_file_as_dict(self):
         """Load the config file into self.config, with recursive loading."""
         def get_config():
             """Unnecessary now, but a deprecation warning is more trouble than it's worth."""
             return self.config
-        
+
         namespace = dict(
             c=self.config,
             load_subconfig=self.load_subconfig,
             get_config=get_config,
             __file__=self.full_filename,
         )
-        fs_encoding = sys.getfilesystemencoding() or 'ascii'
-        conf_filename = self.full_filename.encode(fs_encoding)
-        py3compat.execfile(conf_filename, namespace)
+        conf_filename = self.full_filename
+        with open(conf_filename, 'rb') as f:
+            exec(compile(f.read(), conf_filename, 'exec'), namespace, namespace)
 
 
 class CommandLineConfigLoader(ConfigLoader):
@@ -496,215 +655,144 @@ class CommandLineConfigLoader(ConfigLoader):
     here.
     """
 
-    def _exec_config_str(self, lhs, rhs):
+    def _exec_config_str(self, lhs, rhs, trait=None):
         """execute self.config.<lhs> = <rhs>
-        
-        * expands ~ with expanduser
-        * tries to assign with literal_eval, otherwise assigns with just the string,
-          allowing `--C.a=foobar` and `--C.a="foobar"` to be equivalent.  *Not*
-          equivalent are `--C.a=4` and `--C.a='4'`.
-        """
-        rhs = os.path.expanduser(rhs)
-        try:
-            # Try to see if regular Python syntax will work. This
-            # won't handle strings as the quote marks are removed
-            # by the system shell.
-            value = literal_eval(rhs)
-        except (NameError, SyntaxError, ValueError):
-            # This case happens if the rhs is a string.
-            value = rhs
 
-        exec(u'self.config.%s = value' % lhs)
+        * expands ~ with expanduser
+        * interprets value with trait if available
+        """
+        value = rhs
+        if isinstance(value, DeferredConfig):
+            if trait:
+                # trait available, reify config immediately
+                value = value.get_value(trait)
+            elif isinstance(rhs, DeferredConfigList) and len(rhs) == 1:
+                # single item, make it a deferred str
+                value = DeferredConfigString(os.path.expanduser(rhs[0]))
+        else:
+            if trait:
+                value = trait.from_string(value)
+            else:
+                value = DeferredConfigString(value)
+
+        *path, key = lhs.split(".")
+        section = self.config
+        for part in path:
+            section = section[part]
+        section[key] = value
+        return
 
     def _load_flag(self, cfg):
         """update self.config from a flag, which can be a dict or Config"""
         if isinstance(cfg, (dict, Config)):
             # don't clobber whole config sections, update
             # each section from config:
-            for sec,c in cfg.items():
+            for sec, c in cfg.items():
                 self.config[sec].update(c)
         else:
             raise TypeError("Invalid flag: %r" % cfg)
 
-# raw --identifier=value pattern
-# but *also* accept '-' as wordsep, for aliases
-# accepts:  --foo=a
-#           --Class.trait=value
-#           --alias-name=value
-# rejects:  -foo=value
-#           --foo
-#           --Class.trait
-kv_pattern = re.compile(r'\-\-[A-Za-z][\w\-]*(\.[\w\-]+)*\=.*')
+# match --Class.trait keys for argparse
+# matches:
+# --Class.trait
+# --x
+# -x
 
-# just flags, no assignments, with two *or one* leading '-'
-# accepts:  --foo
-#           -foo-bar-again
-# rejects:  --anything=anything
-#           --two.word
+class_trait_opt_pattern = re.compile(r'^\-?\-[A-Za-z][\w]*(\.[\w]+)*$')
 
-flag_pattern = re.compile(r'\-\-?\w+[\-\w]*$')
+_DOT_REPLACEMENT = "__DOT__"
+_DASH_REPLACEMENT = "__DASH__"
 
-class KeyValueConfigLoader(CommandLineConfigLoader):
-    """A config loader that loads key value pairs from the command line.
 
-    This allows command line options to be gives in the following form::
+class _KVAction(argparse.Action):
+    """Custom argparse action for handling --Class.trait=x
 
-        ipython --profile="foo" --InteractiveShell.autocall=False
+    Always
     """
-
-    def __init__(self, argv=None, aliases=None, flags=None, **kw):
-        """Create a key value pair config loader.
-
-        Parameters
-        ----------
-        argv : list
-            A list that has the form of sys.argv[1:] which has unicode
-            elements of the form u"key=value". If this is None (default),
-            then sys.argv[1:] will be used.
-        aliases : dict
-            A dict of aliases for configurable traits.
-            Keys are the short aliases, Values are the resolved trait.
-            Of the form: `{'alias' : 'Configurable.trait'}`
-        flags : dict
-            A dict of flags, keyed by str name. Vaues can be Config objects,
-            dicts, or "key=value" strings.  If Config or dict, when the flag
-            is triggered, The flag is loaded as `self.config.update(m)`.
-
-        Returns
-        -------
-        config : Config
-            The resulting Config object.
-
-        Examples
-        --------
-
-            >>> from traitlets.config.loader import KeyValueConfigLoader
-            >>> cl = KeyValueConfigLoader()
-            >>> d = cl.load_config(["--A.name='brian'","--B.number=0"])
-            >>> sorted(d.items())
-            [('A', {'name': 'brian'}), ('B', {'number': 0})]
-        """
-        super(KeyValueConfigLoader, self).__init__(**kw)
-        if argv is None:
-            argv = sys.argv[1:]
-        self.argv = argv
-        self.aliases = aliases or {}
-        self.flags = flags or {}
+    def __call__(self, parser, namespace, values, option_string=None):
+        if isinstance(values, str):
+            values = [values]
+        values = ["-" if v is _DASH_REPLACEMENT else v for v in values]
+        items = getattr(namespace, self.dest, None)
+        if items is None:
+            items = DeferredConfigList()
+        else:
+            items = DeferredConfigList(items)
+        items.extend(values)
+        setattr(namespace, self.dest, items)
 
 
-    def clear(self):
-        super(KeyValueConfigLoader, self).clear()
-        self.extra_args = []
+class _DefaultOptionDict(dict):
+    """Like the default options dict
+
+    but acts as if all --Class.trait options are predefined
+    """
+    def _add_kv_action(self, key):
+        self[key] = _KVAction(
+            option_strings=[key],
+            dest=key.lstrip("-").replace(".", _DOT_REPLACEMENT),
+            # use metavar for display purposes
+            metavar=key.lstrip("-"),
+        )
+
+    def __contains__(self, key):
+        if '=' in key:
+            return False
+        if super().__contains__(key):
+            return True
+
+        if key.startswith("-") and class_trait_opt_pattern.match(key):
+            self._add_kv_action(key)
+            return True
+        return False
+
+    def __getitem__(self, key):
+        if key in self:
+            return super().__getitem__(key)
+        else:
+            raise KeyError(key)
+
+    def get(self, key, default=None):
+        try:
+            return self[key]
+        except KeyError:
+            return default
 
 
-    def _decode_argv(self, argv, enc=None):
-        """decode argv if bytes, using stdin.encoding, falling back on default enc"""
-        uargv = []
-        if enc is None:
-            enc = DEFAULT_ENCODING
-        for arg in argv:
-            if not isinstance(arg, text_type):
-                # only decode if not already decoded
-                arg = arg.decode(enc)
-            uargv.append(arg)
-        return uargv
+class _KVArgParser(argparse.ArgumentParser):
+    """subclass of ArgumentParser where any --Class.trait option is implicitly defined"""
+    def parse_known_args(self, args=None, namespace=None):
+        # must be done immediately prior to parsing because if we do it in init,
+        # registration of explicit actions via parser.add_option will fail during setup
+        for container in (self, self._optionals):
+            container._option_string_actions = _DefaultOptionDict(
+                container._option_string_actions)
+        return super().parse_known_args(args, namespace)
 
 
-    def load_config(self, argv=None, aliases=None, flags=None):
-        """Parse the configuration and generate the Config object.
-
-        After loading, any arguments that are not key-value or
-        flags will be stored in self.extra_args - a list of
-        unparsed command-line arguments.  This is used for
-        arguments such as input files or subcommands.
-
-        Parameters
-        ----------
-        argv : list, optional
-            A list that has the form of sys.argv[1:] which has unicode
-            elements of the form u"key=value". If this is None (default),
-            then self.argv will be used.
-        aliases : dict
-            A dict of aliases for configurable traits.
-            Keys are the short aliases, Values are the resolved trait.
-            Of the form: `{'alias' : 'Configurable.trait'}`
-        flags : dict
-            A dict of flags, keyed by str name. Values can be Config objects
-            or dicts.  When the flag is triggered, The config is loaded as
-            `self.config.update(cfg)`.
-        """
-        self.clear()
-        if argv is None:
-            argv = self.argv
-        if aliases is None:
-            aliases = self.aliases
-        if flags is None:
-            flags = self.flags
-
-        # ensure argv is a list of unicode strings:
-        uargv = self._decode_argv(argv)
-        for idx,raw in enumerate(uargv):
-            # strip leading '-'
-            item = raw.lstrip('-')
-
-            if raw == '--':
-                # don't parse arguments after '--'
-                # this is useful for relaying arguments to scripts, e.g.
-                # ipython -i foo.py --matplotlib=qt -- args after '--' go-to-foo.py
-                self.extra_args.extend(uargv[idx+1:])
-                break
-
-            if kv_pattern.match(raw):
-                lhs,rhs = item.split('=',1)
-                # Substitute longnames for aliases.
-                if lhs in aliases:
-                    lhs = aliases[lhs]
-                if '.' not in lhs:
-                    # probably a mistyped alias, but not technically illegal
-                    self.log.warning("Unrecognized alias: '%s', it will probably have no effect.", raw)
-                try:
-                    self._exec_config_str(lhs, rhs)
-                except Exception:
-                    raise ArgumentError("Invalid argument: '%s'" % raw)
-
-            elif flag_pattern.match(raw):
-                if item in flags:
-                    cfg,help = flags[item]
-                    self._load_flag(cfg)
-                else:
-                    raise ArgumentError("Unrecognized flag: '%s'"%raw)
-            elif raw.startswith('-'):
-                kv = '--'+item
-                if kv_pattern.match(kv):
-                    raise ArgumentError("Invalid argument: '%s', did you mean '%s'?"%(raw, kv))
-                else:
-                    raise ArgumentError("Invalid argument: '%s'"%raw)
-            else:
-                # keep all args that aren't valid in a list,
-                # in case our parent knows what to do with them.
-                self.extra_args.append(item)
-        return self.config
-
 class ArgParseConfigLoader(CommandLineConfigLoader):
     """A loader that uses the argparse module to load from the command line."""
 
-    def __init__(self, argv=None, aliases=None, flags=None, log=None,  *parser_args, **parser_kw):
+    parser_class = ArgumentParser
+
+    def __init__(self, argv=None, aliases=None, flags=None, log=None, classes=(),
+                 *parser_args, **parser_kw):
         """Create a config loader for use with argparse.
 
         Parameters
         ----------
-
+        classes : optional, list
+            The classes to scan for *container* config-traits and decide
+            for their "multiplicity" when adding them as *argparse* arguments.
         argv : optional, list
-          If given, used to read command-line arguments from, otherwise
-          sys.argv[1:] is used.
-
+            If given, used to read command-line arguments from, otherwise
+            sys.argv[1:] is used.
         parser_args : tuple
-          A tuple of positional arguments that will be passed to the
-          constructor of :class:`argparse.ArgumentParser`.
-
+            A tuple of positional arguments that will be passed to the
+            constructor of :class:`argparse.ArgumentParser`.
         parser_kw : dict
-          A tuple of keyword arguments that will be passed to the
-          constructor of :class:`argparse.ArgumentParser`.
+            A tuple of keyword arguments that will be passed to the
+            constructor of :class:`argparse.ArgumentParser`.
 
         Returns
         -------
@@ -718,6 +806,7 @@ class ArgParseConfigLoader(CommandLineConfigLoader):
         self.argv = argv
         self.aliases = aliases or {}
         self.flags = flags or {}
+        self.classes = classes
 
         self.parser_args = parser_args
         self.version = parser_kw.pop("version", None)
@@ -725,24 +814,36 @@ class ArgParseConfigLoader(CommandLineConfigLoader):
         kwargs.update(parser_kw)
         self.parser_kw = kwargs
 
-    def load_config(self, argv=None, aliases=None, flags=None):
+    def load_config(self, argv=None, aliases=None, flags=_deprecated, classes=None):
         """Parse command line arguments and return as a Config object.
 
         Parameters
         ----------
+        argv : optional, list
+            If given, a list with the structure of sys.argv[1:] to parse
+            arguments from. If not given, the instance's self.argv attribute
+            (given at construction time) is used.
+        flags
+            Deprecated in traitlets 5.0, instanciate the config loader with the flags.
+
+        """
+
+        if flags is not _deprecated:
+            warnings.warn(
+                "The `flag` argument to load_config is deprecated since Traitlets "
+                f"5.0 and will be ignored, pass flags the `{type(self)}` constructor.",
+                DeprecationWarning,
+                stacklevel=2,
+            )
 
-        args : optional, list
-          If given, a list with the structure of sys.argv[1:] to parse
-          arguments from. If not given, the instance's self.argv attribute
-          (given at construction time) is used."""
         self.clear()
         if argv is None:
             argv = self.argv
-        if aliases is None:
-            aliases = self.aliases
-        if flags is None:
-            flags = self.flags
-        self._create_parser(aliases, flags)
+        if aliases is not None:
+            self.aliases = aliases
+        if classes is not None:
+            self.classes = classes
+        self._create_parser()
         self._parse_args(argv)
         self._convert_to_config()
         return self.config
@@ -753,91 +854,219 @@ class ArgParseConfigLoader(CommandLineConfigLoader):
         else:
             return []
 
-    def _create_parser(self, aliases=None, flags=None):
-        self.parser = ArgumentParser(*self.parser_args, **self.parser_kw)
-        self._add_arguments(aliases, flags)
+    def _create_parser(self):
+        self.parser = self.parser_class(*self.parser_args, **self.parser_kw)
+        self._add_arguments(self.aliases, self.flags, self.classes)
 
-    def _add_arguments(self, aliases=None, flags=None):
+    def _add_arguments(self, aliases, flags, classes):
         raise NotImplementedError("subclasses must implement _add_arguments")
 
     def _parse_args(self, args):
         """self.parser->self.parsed_data"""
         # decode sys.argv to support unicode command-line options
         enc = DEFAULT_ENCODING
-        uargs = [py3compat.cast_unicode(a, enc) for a in args]
-        self.parsed_data, self.extra_args = self.parser.parse_known_args(uargs)
+        uargs = [cast_unicode(a, enc) for a in args]
+
+        unpacked_aliases = {}
+        if self.aliases:
+            unpacked_aliases = {}
+            for alias, alias_target in self.aliases.items():
+                if alias in self.flags:
+                    continue
+                if not isinstance(alias, tuple):
+                    short_alias, alias = alias, None
+                else:
+                    short_alias, alias = alias
+                for al in (short_alias, alias):
+                    if al is None:
+                        continue
+                    if len(al) == 1:
+                        unpacked_aliases["-" + al] = "--" + alias_target
+                    unpacked_aliases["--" + al] = "--" + alias_target
+
+        def _replace(arg):
+            if arg == "-":
+                return _DASH_REPLACEMENT
+            for k, v in unpacked_aliases.items():
+                if arg == k:
+                    return v
+                if arg.startswith(k + "="):
+                    return v + "=" + arg[len(k) + 1:]
+            return arg
+
+        if '--' in uargs:
+            idx = uargs.index('--')
+            extra_args = uargs[idx+1:]
+            to_parse = uargs[:idx]
+        else:
+            extra_args = []
+            to_parse = uargs
+        to_parse = [_replace(a) for a in to_parse]
+
+        self.parsed_data = self.parser.parse_args(to_parse)
+        self.extra_args = extra_args
 
     def _convert_to_config(self):
         """self.parsed_data->self.config"""
         for k, v in vars(self.parsed_data).items():
-            exec("self.config.%s = v"%k, locals(), globals())
+            *path, key = k.split(".")
+            section = self.config
+            for p in path:
+                section = section[p]
+            setattr(section, key, v)
+
+
+class _FlagAction(argparse.Action):
+    """ArgParse action to handle a flag"""
+    def __init__(self, *args, **kwargs):
+        self.flag = kwargs.pop('flag')
+        self.alias = kwargs.pop('alias', None)
+        kwargs['const'] = Undefined
+        if not self.alias:
+            kwargs['nargs'] = 0
+        super(_FlagAction, self).__init__(*args, **kwargs)
+
+    def __call__(self, parser, namespace, values, option_string=None):
+        if self.nargs == 0 or values is Undefined:
+            namespace._flags.append(self.flag)
+        else:
+            setattr(namespace, self.alias, values)
+
 
 class KVArgParseConfigLoader(ArgParseConfigLoader):
     """A config loader that loads aliases and flags with argparse,
-    but will use KVLoader for the rest.  This allows better parsing
-    of common args, such as `ipython -c 'print 5'`, but still gets
-    arbitrary config with `ipython --InteractiveShell.use_readline=False`"""
 
-    def _add_arguments(self, aliases=None, flags=None):
-        self.alias_flags = {}
-        # print aliases, flags
-        if aliases is None:
-            aliases = self.aliases
-        if flags is None:
-            flags = self.flags
+    as well as arbitrary --Class.trait value
+    """
+
+    parser_class = _KVArgParser
+
+    def _add_arguments(self, aliases, flags, classes):
+        alias_flags = {}
         paa = self.parser.add_argument
-        for key,value in aliases.items():
-            if key in flags:
-                # flags
-                nargs = '?'
-            else:
-                nargs = None
-            if len(key) is 1:
-                paa('-'+key, '--'+key, type=text_type, dest=value, nargs=nargs)
-            else:
-                paa('--'+key, type=text_type, dest=value, nargs=nargs)
-        for key, (value, help) in flags.items():
-            if key in self.aliases:
-                #
-                self.alias_flags[self.aliases[key]] = value
-                continue
-            if len(key) is 1:
-                paa('-'+key, '--'+key, action='append_const', dest='_flags', const=value)
-            else:
-                paa('--'+key, action='append_const', dest='_flags', const=value)
+        self.parser.set_defaults(_flags=[])
+        paa("extra_args", nargs="*")
+
+        ## An index of all container traits collected::
+        #
+        #     { <traitname>: (<trait>, <argparse-kwds>) }
+        #
+        #  Used to add the correct type into the `config` tree.
+        #  Used also for aliases, not to re-collect them.
+        self.argparse_traits = argparse_traits = {}
+        for cls in classes:
+            for traitname, trait in cls.class_traits(config=True).items():
+                argname = '%s.%s' % (cls.__name__, traitname)
+                argparse_kwds = {'type': str}
+                if isinstance(trait, (Container, Dict)):
+                    multiplicity = trait.metadata.get('multiplicity', 'append')
+                    if multiplicity == 'append':
+                        argparse_kwds['action'] = multiplicity
+                    else:
+                        argparse_kwds['nargs'] = multiplicity
+                argparse_traits[argname] = (trait, argparse_kwds)
+
+        for keys, (value, _) in flags.items():
+            if not isinstance(keys, tuple):
+                keys = (keys,)
+            for key in keys:
+                if key in aliases:
+                    alias_flags[aliases[key]] = value
+                    continue
+                keys = ('-' + key, '--' + key) if len(key) == 1 else ('--' + key,)
+                paa(*keys, action=_FlagAction, flag=value)
+
+        for keys, traitname in aliases.items():
+            if not isinstance(keys, tuple):
+                keys = (keys,)
+
+            for key in keys:
+                argparse_kwds = {
+                    'type': str,
+                    'dest': traitname.replace(".", _DOT_REPLACEMENT),
+                    'metavar': traitname,
+                }
+                if traitname in argparse_traits:
+                    argparse_kwds.update(argparse_traits[traitname][1])
+                    if 'action' in argparse_kwds and traitname in alias_flags:
+                        # flag sets 'action', so can't have flag & alias with custom action
+                        # on the same name
+                        raise ArgumentError(
+                            "The alias `%s` for the 'append' sequence "
+                            "config-trait `%s` cannot be also a flag!'"
+                            % (key, traitname))
+                if traitname in alias_flags:
+                    # alias and flag.
+                    # when called with 0 args: flag
+                    # when called with >= 1: alias
+                    argparse_kwds.setdefault('nargs', '?')
+                    argparse_kwds['action'] = _FlagAction
+                    argparse_kwds['flag'] = alias_flags[traitname]
+                    argparse_kwds['alias'] = traitname
+                keys = ('-' + key, '--' + key) if len(key) == 1 else ('--'+ key,)
+                paa(*keys, **argparse_kwds)
 
     def _convert_to_config(self):
         """self.parsed_data->self.config, parse unrecognized extra args via KVLoader."""
-        # remove subconfigs list from namespace before transforming the Namespace
-        if '_flags' in self.parsed_data:
-            subcs = self.parsed_data._flags
-            del self.parsed_data._flags
-        else:
-            subcs = []
+        extra_args = self.extra_args
 
-        for k, v in vars(self.parsed_data).items():
-            if v is None:
-                # it was a flag that shares the name of an alias
-                subcs.append(self.alias_flags[k])
-            else:
-                # eval the KV assignment
-                self._exec_config_str(k, v)
+        for lhs, rhs in vars(self.parsed_data).items():
+            if lhs == "extra_args":
+                self.extra_args = ["-" if a == _DASH_REPLACEMENT else a for a in rhs] + extra_args
+                continue
+            elif lhs == '_flags':
+                # _flags will be handled later
+                continue
 
-        for subc in subcs:
+            lhs = lhs.replace(_DOT_REPLACEMENT, ".")
+            if '.' not in lhs:
+                # probably a mistyped alias, but not technically illegal
+                self.log.warning("Unrecognized alias: '%s', it will have no effect.", lhs)
+                trait = None
+
+            if isinstance(rhs, list):
+                rhs = DeferredConfigList(rhs)
+            elif isinstance(rhs, str):
+                rhs = DeferredConfigString(rhs)
+
+            trait = self.argparse_traits.get(lhs)
+            if trait:
+                trait = trait[0]
+
+            # eval the KV assignment
+            try:
+                self._exec_config_str(lhs, rhs, trait)
+            except Exception as e:
+                # cast deferred to nicer repr for the error
+                # DeferredList->list, etc
+                if isinstance(rhs, DeferredConfig):
+                    rhs = rhs._super_repr()
+                raise ArgumentError(f"Error loading argument {lhs}={rhs}, {e}")
+
+        for subc in self.parsed_data._flags:
             self._load_flag(subc)
 
-        if self.extra_args:
-            sub_parser = KeyValueConfigLoader(log=self.log)
-            sub_parser.load_config(self.extra_args)
-            self.config.merge(sub_parser.config)
-            self.extra_args = sub_parser.extra_args
+
+class KeyValueConfigLoader(KVArgParseConfigLoader):
+    """Deprecated in traitlets 5.0
+
+    Use KVArgParseConfigLoader
+    """
+    def __init__(self, *args, **kwargs):
+        warnings.warn(
+            "KeyValueConfigLoader is deprecated since Traitlets 5.0."
+            " Use KVArgParseConfigLoader instead.",
+            DeprecationWarning,
+            stacklevel=2,
+        )
+        super().__init__(*args, **kwargs)
 
 
 def load_pyconfig_files(config_files, path):
     """Load multiple Python config files, merging each of them in turn.
 
     Parameters
-    ==========
+    ----------
     config_files : list of str
         List of config files names to load and merge into the config.
     path : unicode
